from typing import Any, NamedTuple

from .._models import Entity, create_model
from .object import Species


def _create_species_model(std_id="scientific_name"):
    df = Species().df.reset_index()
    SpeciesData = create_model("SpeciesData", __base__=Entity)
    for i in df.index:
        entry = {
            "name": df.loc[i][std_id],
        }
        entry.update({col: df.loc[i][col] for col in df.columns})
        SpeciesData.add_fields(**{df.loc[i][std_id]: (Entry, Entry(**entry))})
    return SpeciesData


class Entry(NamedTuple):
    name: str  # this is the value of std_id, aka scientific_name
    display_name: str
    scientific_name: str
    common_name: str
    taxon_id: int
    assembly: str
    accession: str
    release: int
    short_name: str


SpeciesDataModel: Any = _create_species_model()


class SpeciesModel(SpeciesDataModel):
    def __call__(self, **kwargs):
        return Species(**kwargs)


species = SpeciesModel(**{"name": "species", "std_id": "scientific_name"})
"""Access point to species.

`species` -> a static class containing species as fields
`species()` -> an instance of bionty.species.Species
"""



from functools import wraps
from pathlib import Path
from typing import Union

import pandas as pd


def check_datasetdir_exists(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        settings.datasetdir.mkdir(exist_ok=True)
        return f(*args, **kwargs)

    return wrapper


def format_into_dataframe(f):
    @wraps(f)
    def dataframe(*args, **kwargs) -> pd.DataFrame:
        # Check if the first argument is self
        idx = 0 if _is_function(args[0]) else 1

        if isinstance(args[idx], pd.DataFrame):
            df = args[idx]
            reformat = False
        else:
            df = pd.DataFrame(index=[d for d in args[idx]])
            reformat = True

        args_new = list(args)
        args_new[idx] = df
        return f(*args_new, _reformat=reformat, **kwargs)

    return dataframe


def _is_function(func) -> bool:
    """Checks if input is a function rather than an instance of class.

    TODO: Is this hacky?
    """
    return hasattr(func, "__name__")


class Settings:
    def __init__(self, datasetdir: Union[str, Path] = Path("./data/")):
        self._datasetdir = datasetdir

    @property
    def datasetdir(self):
        """Directory for datasets."""
        return self._datasetdir

    @datasetdir.setter
    def datasetdir(self, datasetdir: Union[str, Path]):
        self._datasetdir = Path(datasetdir).resolve()


settings = Settings()



"""Bionty: mapping and standardizing biological entities.

Import the package::

   import bionty as bt

This is the complete API reference:

Entity collections:

.. autosummary::
   :toctree: .

   species

Entity classes:

.. autosummary::
   :toctree: .

   Celltype
   Disease
   Gene
   Protein
   Species
   Tissue

Other tools:

.. autosummary::
   :toctree: .

   Ontology
"""

from . import _version

__version__ = "0.1a1"

# bio entities
from .gene import Gene
from .protein import Protein
from .species import species, Species
from .celltype import Celltype
from .disease import Disease
from .tissue import Tissue

# tools
from ._normalize import NormalizeColumns
from ._ontology import Ontology



from functools import cached_property
from pathlib import Path

import pandas as pd

HERE = Path(__file__).parent
SPECIES_FILENAME = HERE / "tables/Species.csv"


class Species:
    """Species class."""

    def __init__(self, common_name="human") -> None:
        self._std_name = common_name

    @cached_property
    def df(self):
        self._df = pd.read_csv(SPECIES_FILENAME, header=0, index_col=0)
        return self._df

    @property
    def std_id(self):
        """common_name is the standardized id for species."""
        return "display_name"

    @property
    def std_name(self):
        """Value of the .std_id."""
        return self._std_name

    @property
    def fields(self):
        return self.df.columns.tolist()

    def search(self, field: str):
        """Search species fields based on .std_id.

        Args:
            field: one of .fields
        Returns:
            value of a field

        e.g.
        'common_name': 'human'
        'scientific_name': 'homo_sapiens'
        'short_name': 'hsapiens'
        'taxon_id': 9606
        'assembly': 'GRCh38.p13'

        """
        return self.df[[field]].to_dict()[field][self.std_name]
